import numpy as np
from matplotlib import pyplot
import matplotlib.pyplot as plt
from scipy.stats import norm 
import statistics 

#%% Orstein-Uhlenbeck
class Model:
#    """Stochastic model constants."""
    THETA = 0.7
    MU = 1.5
    SIGMA = 0.06

def mu(y: float, _t: float) -> float:
#    """Implement the Ornstein–Uhlenbeck mu."""
    return Model.THETA * (Model.MU - y)

def sigma(_y: float, _t: float) -> float:
#    Implement the Ornstein–Uhlenbeck sigma
    return Model.SIGMA

def dW(delta_t: float) -> float:
#    """Sample a random number at each call."""
    return np.random.randn()*np.sqrt(delta_t)

def run_simulation():
#    """ Return the result of one full simulation."""
    T_INIT = 3
    T_END = 7
    N = 1000  # Compute at 1000 grid points
    DT = float(T_END - T_INIT) / N
    TS = np.arange(T_INIT, T_END + DT, DT)
    assert TS.size == N + 1

    Y_INIT = 0

    ys = np.zeros(TS.size)
    ys[0] = Y_INIT
    for i in range(1, TS.size):
        t = T_INIT + (i - 1) * DT
        y = ys[i - 1]
        ys[i] = y + mu(y, t) * DT + sigma(y, t) * dW(DT)

    return TS, ys


def plot_simulations(num_sims: int):
#    """ Plot several simulations in one image."""
    for _ in range(num_sims):
        plt.plot(*run_simulation())

    plt.xlabel("time")
    plt.ylabel("y")
    plt.show()


NUM_SIMS = 5
plot_simulations(NUM_SIMS)
    
#%% Ej 8

def model_diffusion(t: float, x: float):
        return np.exp(x - t/2)

def run_simulation_ej_8():
#    """ Return the result of one full simulation."""
    T_INIT = 0
    T_END = 1
    N = 1000  # Compute at 1000 grid points
    DT = float(T_END - T_INIT) / N
    TS = np.arange(T_INIT, T_END + DT, DT)
    Y_INIT = 0

    ys = np.zeros(TS.size)
    ys[0] = Y_INIT
    W = 0
    for i in range(1, TS.size):
        t = T_INIT + (i - 1) * DT
        ddW = dW(DT)
        y = ys[i - 1]
        ys[i] = y + model_diffusion(t, W)*ddW
        W = W + ddW
    return TS, ys


def plot_simulations_ej_8(num_sims: int):
#    """ Plot several simulations in one image."""
    y_final = np.zeros(num_sims)
    fig = pyplot.figure()
    fig.set_figheight(8)
    fig.set_figwidth(12)
    for k in range(num_sims):
        [T, y] = run_simulation_ej_8()
        plt.plot(T,y)
        y_final[k] = y[-1]
    plt.xlabel("time")
    plt.ylabel("y") 
    plt.show()
    return y_final



nsims = 10000
y_final = plot_simulations_ej_8(nsims)

#%% part b

values = np.zeros(nsims)
for k in range(nsims):
    values[k] = np.exp(dW(1) - 1/2) - 1

   
print('Mean_Int', np.mean(y_final))
print('variance_Int', np.var(y_final))
     
print('Mean_J', np.mean(values))
print('variance_J', np.var(values))
  
fig = pyplot.figure()
fig.set_figheight(8)
fig.set_figwidth(12)
pyplot.hist(y_final, bins= np.int(np.sqrt(nsims)+ 1), density = True, alpha=0.8, label='Integral'),
pyplot.hist(values, bins= np.int(np.sqrt(nsims)+ 1), density = True, alpha=0.3, label='J'),
pyplot.legend(loc='upper right')
pyplot.show()


#%% Ej 9

def model_ej_9(t: float, x: float):
        return np.exp(t)

def run_simulation_ej_9():
#    """ Return the result of one full simulation."""
    T_INIT = 0
    T_END = 1
    N = 1000  # Compute at 1000 grid points
    DT = float(T_END - T_INIT) / N
    TS = np.arange(T_INIT, T_END + DT, DT)
    Y_INIT = 0

    ys = np.zeros(TS.size)
    ys[0] = Y_INIT
    W = 0
    for i in range(1, TS.size):
        t = T_INIT + (i - 1) * DT
        ddW = dW(DT)
        y = ys[i - 1]
        ys[i] = y + model_ej_9(t, W)*ddW
        W = W + ddW
    return TS, ys


def plot_simulations_ej_9(num_sims: int):
#    """ Plot several simulations in one image."""
    y_final = np.zeros(num_sims)
    fig = pyplot.figure()
    fig.set_figheight(8)
    fig.set_figwidth(12)

    for k in range(num_sims):
        [T, y] = run_simulation_ej_9()
        plt.plot(T,y,'y')
        y_final[k] = y[-1]
        if k==num_sims-1:
            plt.plot(T,np.sqrt((np.exp(2*T)-1)/2),'k')
            plt.plot(T,-np.sqrt((np.exp(2*T)-1)/2),'k')
            plt.plot(T,2*np.sqrt((np.exp(2*T)-1)/2),'b')
            plt.plot(T,-2*np.sqrt((np.exp(2*T)-1)/2),'b')
            plt.plot(T,3*np.sqrt((np.exp(2*T)-1)/2),'c')
            plt.plot(T,-3*np.sqrt((np.exp(2*T)-1)/2),'c')
    plt.xlabel("time")
    plt.ylabel("y")
    plt.show()
    return y_final



nsims = 100
y_final = plot_simulations_ej_9(nsims)


print('Mean', np.mean(y_final))
print('variance', np.var(y_final))
print('th_var', (np.exp(2)-1)/2)

#%% Histogram

  
fig = pyplot.figure()
fig.set_figheight(8)
fig.set_figwidth(12)
pyplot.hist(y_final, bins= np.int(np.sqrt(nsims)+ 1), density = True, alpha=0.8, label='Integral'),
x_axis = np.arange(-7, 7, 0.01) 
pyplot.plot(x_axis, norm.pdf(x_axis,np.mean(y_final) ,np.sqrt(np.var(y_final))))
pyplot.legend(loc='upper right')
pyplot.show()
